@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2000--2013  The R Core Team
+ *  Copyright (C) 2000--2022  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,12 +14,17 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, a copy is available at
- *  http://www.r-project.org/Licenses/
+ *  https://www.R-project.org/Licenses/
  */
 
-#include "tcltk.h" /* declarations of our `public' interface */
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+#define NO_NLS
+#include <Defn.h>
+#include <R_ext/RS.h> // for R_Calloc
 
-#include <R.h>
+#include "tcltk.h" /* declarations of our `public' interface */
 #include <stdlib.h>
 
 #ifdef ENABLE_NLS
@@ -29,8 +34,7 @@
 #define _(String) (String)
 #endif
 
-// From Defn.f
-extern void  R_Busy(int);
+Tcl_Interp *RTcl_interp;
 
 static void RTcl_dec_refcount(SEXP R_tclobj)
 {
@@ -74,6 +78,7 @@ static int R_eval(ClientData clientData,
 	int n = length(expr);
 	for(i = 0 ; i < n ; i++)
 	    ans = eval(VECTOR_ELT(expr, i), R_GlobalEnv);
+	PROTECT(ans);
 	R_Busy(0);
     }
 
@@ -81,7 +86,7 @@ static int R_eval(ClientData clientData,
     if (inherits(ans, "tclObj"))
 	    Tcl_SetObjResult(interp, (Tcl_Obj*) R_ExternalPtrAddr(ans));
 
-    UNPROTECT(2);
+    UNPROTECT(3);
     return TCL_OK;
 }
 
@@ -106,6 +111,8 @@ static int R_call(ClientData clientData,
     SEXP expr, alist, ans;
     void *fun;
 
+    SEXP s_try = install("try");
+
     alist = R_NilValue;
     for (i = argc - 1 ; i > 1 ; i--){
 	PROTECT(alist);
@@ -116,16 +123,17 @@ static int R_call(ClientData clientData,
     sscanf(argv[1], "%p", &fun);
 
     expr = LCONS( (SEXP)fun, alist);
-    expr = LCONS(install("try"), LCONS(expr, R_NilValue));
+    PROTECT(expr = LCONS(s_try, LCONS(expr, R_NilValue)));
 
     R_Busy(1);
-    ans = eval(expr, R_GlobalEnv);
+    PROTECT(ans = eval(expr, R_GlobalEnv));
     R_Busy(0);
 	
     /* If return value is of class tclObj, use as Tcl result */
     if (inherits(ans, "tclObj"))
 	Tcl_SetObjResult(interp, (Tcl_Obj*) R_ExternalPtrAddr(ans));
 
+    UNPROTECT(2);
     return TCL_OK;
 }
 
@@ -140,16 +148,19 @@ static int R_call_lang(ClientData clientData,
     sscanf(argv[1], "%p", &expr);
     sscanf(argv[2], "%p", &env);
 
-    expr = LCONS(install("try"), LCONS(expr, R_NilValue));
+    SEXP s_try = install("try");
+    expr = LCONS(s_try, LCONS(expr, R_NilValue));
+    PROTECT((SEXP)expr);
 
     R_Busy(1);
-    ans = eval((SEXP)expr, (SEXP)env);
+    PROTECT(ans = eval((SEXP)expr, (SEXP)env));
     R_Busy(0);
 
     /* If return value is of class tclObj, use as Tcl result */
     if (inherits(ans, "tclObj"))
 	Tcl_SetObjResult(interp, (Tcl_Obj*) R_ExternalPtrAddr(ans));
 
+    UNPROTECT(2);
     return TCL_OK;
 }
 
@@ -224,11 +235,12 @@ SEXP dotTclObjv(SEXP args)
 	const char *s;
 	char *tmp;
 	if (!isNull(nm) && strlen(s = translateChar(STRING_ELT(nm, i)))){
-	    tmp = calloc(strlen(s)+2, sizeof(char));
+	    //  tmp = calloc(strlen(s)+2, sizeof(char));
+	    tmp = R_Calloc(strlen(s)+2, char);
 	    *tmp = '-';
 	    strcpy(tmp+1, s);
 	    objv[objc++] = Tcl_NewStringObj(tmp, -1);
-	    free(tmp);
+	    R_Free(tmp);
 	}
 	if (!isNull(t = VECTOR_ELT(avec, i)))
 	    objv[objc++] = (Tcl_Obj *) R_ExternalPtrAddr(t);
@@ -267,10 +279,15 @@ SEXP RTcl_ObjFromVar(SEXP args)
     Tcl_Obj *tclobj;
     const void *vmax = vmaxget();
 
+    if(!isValidString(CADR(args)))
+	error(_("invalid argument"));
     tclobj = Tcl_GetVar2Ex(RTcl_interp,
                            translateChar(STRING_ELT(CADR(args), 0)),
                            NULL,
                            0);
+    if (tclobj == NULL)
+	/* the variable may have been deleted using "unset" */
+	error(_("no such variable"));
     SEXP res = makeRTclObject(tclobj);
     vmaxset(vmax);
     return res;
@@ -279,6 +296,8 @@ SEXP RTcl_ObjFromVar(SEXP args)
 SEXP RTcl_AssignObjToVar(SEXP args)
 {
     const void *vmax = vmaxget();
+    if(!isValidString(CADR(args)))
+	error(_("invalid argument"));
     Tcl_SetVar2Ex(RTcl_interp,
 		  translateChar(STRING_ELT(CADR(args), 0)),
 		  NULL,
@@ -297,6 +316,8 @@ SEXP RTcl_StringFromObj(SEXP args)
     Tcl_DString s_ds;
     Tcl_Obj *obj;
 
+    if (TYPEOF(CADR(args)) != EXTPTRSXP)
+	error(_("invalid argument"));
     obj = (Tcl_Obj *) R_ExternalPtrAddr(CADR(args));
     if (!obj) error(_("invalid tclObj -- perhaps saved from another session?"));
     Tcl_DStringInit(&s_ds);
@@ -315,6 +336,8 @@ SEXP RTcl_ObjAsCharVector(SEXP args)
     int ret, i;
     SEXP ans;
 
+    if (TYPEOF(CADR(args)) != EXTPTRSXP)
+	error(_("invalid argument"));
     obj = (Tcl_Obj *) R_ExternalPtrAddr(CADR(args));
     if (!obj) error(_("invalid tclObj -- perhaps saved from another session?"));
     ret = Tcl_ListObjGetElements(RTcl_interp, obj, &count, &elem);
@@ -388,6 +411,8 @@ SEXP RTcl_ObjAsDoubleVector(SEXP args)
     double x;
     SEXP ans;
 
+    if (TYPEOF(CADR(args)) != EXTPTRSXP)
+	error(_("invalid argument"));
     obj = (Tcl_Obj *) R_ExternalPtrAddr(CADR(args));
     if (!obj) error(_("invalid tclObj -- perhaps saved from another session?"));
 
@@ -451,6 +476,8 @@ SEXP RTcl_ObjAsIntVector(SEXP args)
     int x;
     SEXP ans;
 
+    if (TYPEOF(CADR(args)) != EXTPTRSXP)
+	error(_("invalid argument"));
     obj = (Tcl_Obj *) R_ExternalPtrAddr(CADR(args));
     if (!obj) error(_("invalid tclObj -- perhaps saved from another session?"));
 
@@ -503,6 +530,8 @@ SEXP RTcl_ObjAsRawVector(SEXP args)
     unsigned char *ret;
     SEXP ans, el;
 
+    if (TYPEOF(CADR(args)) != EXTPTRSXP)
+	error(_("invalid argument"));
     obj = (Tcl_Obj *) R_ExternalPtrAddr(CADR(args));
     if (!obj) error(_("invalid tclObj -- perhaps saved from another session?"));
     ret = Tcl_GetByteArrayFromObj(obj, &nb);
@@ -677,28 +706,30 @@ void tcltk_init(int *TkUp)
     if (code != TCL_OK) error(Tcl_GetStringResult(RTcl_interp));
 
 /* HAVE_AQUA is not really right here.
-   On Mac OS X we might be using Aqua Tcl/Tk or X11 Tcl/Tk, and that
+   On macOS we might be using Aqua Tcl/Tk or X11 Tcl/Tk, and that
    is in principle independent of whether we want quartz() built.
 */
 #if !defined(Win32) && !defined(HAVE_AQUA)
     char *p= getenv("DISPLAY");
-    if(p && p[0])  /* exclude DISPLAY = "" */
+    if(!getenv("R_DONT_USE_TK")) {
+	if(p && p[0])  /* exclude DISPLAY = "" */
 #endif
-    {
-	code = Tk_Init(RTcl_interp);  /* Load Tk into interpreter */
-	if (code != TCL_OK) {
-	    warning(Tcl_GetStringResult(RTcl_interp));
-	} else {
-	    Tcl_StaticPackage(RTcl_interp, "Tk", Tk_Init, Tk_SafeInit);
-
-	    code = Tcl_Eval(RTcl_interp, "wm withdraw .");  /* Hide window */
-	    if (code != TCL_OK) error(Tcl_GetStringResult(RTcl_interp));
-	    *TkUp = 1;
+	{
+	    code = Tk_Init(RTcl_interp);  /* Load Tk into interpreter */
+	    if (code != TCL_OK) {
+		warning(Tcl_GetStringResult(RTcl_interp));
+	    } else {
+		Tcl_StaticPackage(RTcl_interp, "Tk", Tk_Init, Tk_SafeInit);
+		
+		code = Tcl_Eval(RTcl_interp, "wm withdraw .");  /* Hide window */
+		if (code != TCL_OK) error(Tcl_GetStringResult(RTcl_interp));
+		*TkUp = 1;
+	    }
 	}
-    }
 #if !defined(Win32) && !defined(HAVE_AQUA)
-    else
-	warning(_("no DISPLAY variable so Tk is not available"));
+	else
+	    warningcall(R_NilValue, _("no DISPLAY variable so Tk is not available"));
+    }
 #endif
 
     Tcl_CreateCommand(RTcl_interp,
