@@ -1,7 +1,7 @@
 #   File src/library/utils/R/SweaveDrivers.R
-#  Part of the R package, http://www.R-project.org
+#  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2014 The R Core Team
+#  Copyright (C) 1995-2016 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
 #  GNU General Public License for more details.
 #
 #  A copy of the GNU General Public License is available at
-#  http://www.r-project.org/Licenses/
+#  https://www.R-project.org/Licenses/
 
 RweaveLatex <- function()
 {
@@ -36,7 +36,7 @@ RweaveLatexSetup <-
     dots <- list(...)
     if (is.null(output)) {
         prefix.string <- basename(sub(syntax$extension, "", file))
-        output <- paste(prefix.string, "tex", sep = ".")
+        output <- paste0(prefix.string, ".tex")
     } else prefix.string <- basename(sub("\\.tex$", "", output))
 
     if (!quiet) cat("Writing to file ", output, "\n",
@@ -95,21 +95,21 @@ makeRweaveLatexCodeRunner <- function(evalFunc = RweaveEvalWithOpt)
     ## only once.
     function(object, chunk, options) {
         pdf.Swd <- function(name, width, height, ...)
-            grDevices::pdf(file = paste(chunkprefix, "pdf", sep = "."),
+            grDevices::pdf(file = paste0(chunkprefix, ".pdf"),
                            width = width, height = height,
                            version = options$pdf.version,
                            encoding = options$pdf.encoding,
                            compress = options$pdf.compress)
         eps.Swd <- function(name, width, height, ...)
-            grDevices::postscript(file = paste(name, "eps", sep = "."),
+            grDevices::postscript(file = paste0(name, ".eps"),
                                   width = width, height = height,
                                   paper = "special", horizontal = FALSE)
         png.Swd <- function(name, width, height, options, ...)
-            grDevices::png(filename = paste(chunkprefix, "png", sep = "."),
+            grDevices::png(filename = paste0(chunkprefix, ".png"),
                            width = width, height = height,
                            res = options$resolution, units = "in")
         jpeg.Swd <- function(name, width, height, options, ...)
-            grDevices::jpeg(filename = paste(chunkprefix, "jpeg", sep = "."),
+            grDevices::jpeg(filename = paste0(chunkprefix, ".jpeg"),
                             width = width, height = height,
                             res = options$resolution, units = "in")
 
@@ -133,13 +133,26 @@ makeRweaveLatexCodeRunner <- function(evalFunc = RweaveEvalWithOpt)
                 devs <- c(devs, list(jpeg.Swd))
                 devoffs <- c(devoffs, list(grDevices::dev.off))
             }
-            if (nzchar(grd <- options$grdevice)) {
-                devs <- c(devs, list(get(grd, envir = .GlobalEnv)))
-                grdo <- paste(grd, "off", sep = ".")
-                devoffs <- c(devoffs,
-                             if (exists(grdo, envir = .GlobalEnv))
-                                 list(get(grdo, envir = .GlobalEnv))
-                             else list(grDevices::dev.off))
+            if(nzchar(grd <- options$grdevice)) {
+                grdo <- paste0(grd, ".off")
+                if(grepl("::", grd, fixed = TRUE)) {
+                    devs <- c(devs, eval(str2expression(grd)))
+                    devoffs <-
+                        c(devoffs,
+                          if(!inherits(grdo <- tryCatch(eval(str2expression(grdo)), error = identity),
+                                       "error"))
+                              list(grdo)
+                          else
+                              list(grDevices::dev.off))
+                } else {
+                    devs <- c(devs, list(get(grd, envir = .GlobalEnv)))
+                    devoffs <-
+                        c(devoffs,
+                          if(exists(grdo, envir = .GlobalEnv))
+                              list(get(grdo, envir = .GlobalEnv))
+                          else
+                              list(grDevices::dev.off))
+                }
             }
         }
         if (!object$quiet) {
@@ -173,7 +186,7 @@ makeRweaveLatexCodeRunner <- function(evalFunc = RweaveEvalWithOpt)
             ## [x][[1L]] avoids partial matching of x
             chunkout <- object$chunkout[chunkprefix][[1L]]
             if (is.null(chunkout)) {
-                chunkout <- file(paste(chunkprefix, "tex", sep = "."), "w")
+                chunkout <- file(paste0(chunkprefix, ".tex"), "w")
                 if (!is.null(options$label))
                     object$chunkout[[chunkprefix]] <- chunkout
                 if(!grepl(.SweaveValidFilenameRegexp, chunkout))
@@ -207,12 +220,12 @@ makeRweaveLatexCodeRunner <- function(evalFunc = RweaveEvalWithOpt)
                 openSinput <<- TRUE
             }
             leading <- max(leading, 1L) # safety check
-            cat("\n", paste(getOption("prompt"), dce[seq_len(leading)],
-                            sep = "", collapse = "\n"),
+            cat("\n", paste0(getOption("prompt"), dce[seq_len(leading)],
+                             collapse = "\n"),
                 file = chunkout, sep = "")
             if (length(dce) > leading)
-                cat("\n", paste(getOption("continue"), dce[-seq_len(leading)],
-                                sep = "", collapse = "\n"),
+                cat("\n", paste0(getOption("continue"), dce[-seq_len(leading)],
+                                 collapse = "\n"),
                     file = chunkout, sep = "")
             linesout[thisline + seq_along(dce)] <<- srcline
             filenumout[thisline + seq_along(dce)] <<- srcfilenum
@@ -220,19 +233,17 @@ makeRweaveLatexCodeRunner <- function(evalFunc = RweaveEvalWithOpt)
         }
 
         trySrcLines <- function(srcfile, showfrom, showto, ce) {
-	    lines <- tryCatch(suppressWarnings(getSrcLines(srcfile, showfrom, showto)),
-			      error = function(e)e)
-	    if (inherits(lines, "error")) {
-		lines <- if (is.null(ce)) character()
-		else deparse(ce, width.cutoff = 0.75*getOption("width"))
-            }
-            lines
+	    tryCatch(suppressWarnings(getSrcLines(srcfile, showfrom, showto)),
+                     error = function(e) {
+                         if (is.null(ce)) character()
+                         else deparse(ce, width.cutoff = 0.75*getOption("width"))
+                     })
         }
 
         echoComments <- function(showto) {
             if (options$echo && !is.na(lastshown) && lastshown < showto) {
                 dce <- trySrcLines(srcfile, lastshown + 1L, showto, NULL)
-                linedirs <- grepl("^#line ", dce)
+                linedirs <- startsWith(dce, "#line ")
 		dce <- dce[!linedirs]
 		if (length(dce))
                     putSinput(dce, length(dce)) # These are all trailing comments
@@ -281,7 +292,7 @@ makeRweaveLatexCodeRunner <- function(evalFunc = RweaveEvalWithOpt)
                 lastshown <- showto
                 srcline <- srcref[3L]
 
-                linedirs <- grepl("^#line ", dce)
+                linedirs <- startsWith(dce, "#line ")
                 dce <- dce[!linedirs]
                 # Need to reduce leading lines if some were just removed
                 leading <- leading - sum(linedirs[seq_len(leading)])
@@ -437,11 +448,11 @@ RweaveLatexWritedoc <- function(object, chunk)
  	which <- grep(begindoc, chunk)
  	if (length(which)) {
             chunk[which] <- sub(begindoc,
-                                paste("\\\\usepackage{",
-                                      object$styfile,
-                                      "}\n\\\\begin{document}", sep = ""),
+                                paste0("\\\\usepackage{",
+                                       object$styfile,
+                                       "}\n\\\\begin{document}"),
                                 chunk[which])
-            idx <- c(1L:which, which, seq(from = which+1L,
+            idx <- c(1L:which, which, seq.int(from = which+1L,
                      length.out = length(linesout)-which))
             linesout <- linesout[idx]
             filenumout <- filenumout[idx]
@@ -456,12 +467,12 @@ RweaveLatexWritedoc <- function(object, chunk)
                       cmdloc + attr(cmdloc, "match.length") - 1L)
         cmd <- sub(object$syntax$docexpr, "\\1", cmd)
         if (object$options$eval) {
-            val <- tryCatch(as.character(eval(parse(text = cmd), envir = .GlobalEnv)),
+            val <- tryCatch(as.character(eval(str2expression(cmd), envir = .GlobalEnv)),
 		    error = function(e) {
 	               filenum <- attr(chunk, "srcFilenum")[pos[1L]]
                        filename <- attr(chunk, "srcFilenames")[filenum]
                        location <- paste0(basename(filename), ":", attr(chunk, "srclines")[pos[1L]])
-		       stop("at ",location, ", ", conditionMessage(e), call. = FALSE)
+		       stop("at ",location, ", ", conditionMessage(e), domain = NA, call. = FALSE)
 		   })
             ## protect against character(), because sub() will fail
             if (length(val) == 0L) val <- ""
@@ -489,7 +500,7 @@ RweaveLatexWritedoc <- function(object, chunk)
             object$options$label <- "concordance"
             prefix <- RweaveChunkPrefix(object$options)
             object$options$label <- savelabel
-            object$concordfile <- paste(prefix, "tex", sep = ".")
+            object$concordfile <- paste0(prefix, ".tex")
             chunk[pos[1L]] <- sub(object$syntax$docopt,
                                   paste0("\\\\input{", prefix, "}"),
                                   chunk[pos[1L]])
@@ -519,35 +530,18 @@ RweaveLatexFinish <- function(object, error = FALSE)
     if (length(object$chunkout))
         for (con in object$chunkout) close(con)
     if (object$haveconcordance) {
-    	## This output format is subject to change.  Currently it contains
-    	## three or four parts, separated by colons:
-    	## 1.  The output .tex filename
-    	## 2.  The input .Rnw filename
-    	## 3.  Optionally, the starting line number of the output coded as "ofs nn",
-    	##     where nn is the offset to the first output line.  This is omitted if nn is 0.
-    	## 4.  The input line numbers corresponding to each output line.
-    	##     This are compressed using the following simple scheme:
-    	##     The first line number, followed by
-    	##     a run-length encoded diff of the rest of the line numbers.
-        linesout <- object$linesout
-        filenumout <- object$filenumout
-        filenames <- object$srcFilenames[filenumout]
-	if (!is.null(filenames)) {  # Might be NULL if an error occurred
-	    filegps <- rle(filenames)
-	    offset <- 0L
-	    for (i in seq_along(filegps$lengths)) {
-		len <- filegps$lengths[i]
-		inputname <- filegps$values[i]
-		vals <- rle(diff(linesout[offset + seq_len(len)]))
-		vals <- c(linesout[offset + 1L], as.numeric(rbind(vals$lengths, vals$values)))
-		concordance <- paste(strwrap(paste(vals, collapse = " ")), collapse = " %\n")
-		special <- paste0("\\Sconcordance{concordance:", outputname, ":",
-			     inputname, ":",
-			     if (offset) paste0("ofs ", offset, ":") else "",
-			     "%\n",
-			     concordance,"}\n")
-		cat(special, file = object$concordfile, append=offset > 0L)
-		offset <- offset + len
+        concordance <- list(srcLine = object$linesout,
+                            srcFile = object$srcFilenames[object$filenumout],
+        	            offset = 0)
+        class(concordance) <- "Rconcordance"
+        if (!is.null(concordance$srcFile)) { # Might be NULL if an error occurred
+	    concordance <- as.character(concordance, targetfile = outputname)
+	    for (i in seq_along(concordance)) {
+	        special <- paste0("\\Sconcordance{",
+		                  paste(strwrap(concordance[i]),
+		                        collapse = " %\n"),
+		                  "}\n")
+		cat(special, file = object$concordfile, append = (i > 1L))
 	    }
 	}
     }
@@ -633,7 +627,7 @@ RweaveEvalWithOpt <- function (expr, options)
         if (inherits(res, "try-error")) return(res)
         if (options$print || (options$term && res$visible)) {
             if (.isMethodsDispatchOn() && isS4(res$value))
-                methods:::show(res$value) else print(res$value)
+                methods::show(res$value) else print(res$value)
         }
     }
     res
@@ -665,13 +659,13 @@ Rtangle <-  function()
 
 RtangleSetup <-
     function(file, syntax, output = NULL, annotate = TRUE, split = FALSE,
-             quiet = FALSE, ...)
+             quiet = FALSE, drop.evalFALSE = FALSE, ...)
 {
     dots <- list(...)
     if (is.null(output)) {
         prefix.string <- basename(sub(syntax$extension, "", file))
         ## This is odd, since for split = TRUE it uses the engine name.
-        output <- paste(prefix.string, "R", sep = ".")
+        output <- paste0(prefix.string, ".R")
     } else
         prefix.string <- basename(sub("\\.[rsRS]$", "", output))
 
@@ -680,15 +674,14 @@ RtangleSetup <-
         else if (identical(output, "stderr")) output <- stderr()
         else {
             if (!quiet) cat("Writing to file", output, "\n")
-            ## We could at some future point try to write the file in
-            ## 'encoding'.
-            output <- file(output, open = "w")
+            encoding <- attr(file, "encoding")
+            if (encoding %in% c("ASCII", "bytes")) encoding <- ""
+            output <- file(output, open = "w", encoding = encoding)
         }
-        lines <- c(sprintf("R code from vignette source '%s'", file),
-                   if(attr(file, "encoding") != "ASCII")
-                   sprintf("Encoding: %s", localeToCharset()[1L])
-                   )
-        lines <- c(paste("###", lines), "")
+        lines <- c(paste("###",
+                         sprintf("R code from vignette source '%s'",
+                                 file)),
+                   "")
         writeLines(lines, output)
     } else {
         if (!quiet) cat("Writing chunks to files ...\n")
@@ -706,11 +699,25 @@ RtangleSetup <-
     options <- RweaveLatexOptions(options)
 
     list(output = output, annotate = annotate, options = options,
-         chunkout = list(), quiet = quiet, syntax = syntax)
+         chunkout = list(), quiet = quiet, syntax = syntax,
+         drop.evalFALSE = drop.evalFALSE)
 }
 
+.RtangleCodeLabel <- function(chunk) {
+    if(length(lnos <- grep("^#line ", chunk, value = TRUE))) {
+        srclines <- attr(chunk, "srclines")
+        ## srcfilenum <- attr(chunk, "srcFilenum")
+        ## this currently includes the chunk header
+        lno <- if (length(srclines))
+                   paste(min(srclines), max(srclines), sep = "-")
+               else srclines
+        fn <- sub('[^"]*"([^"]+).*', "\\1", lnos[1L])
+        paste(fn, lno, sep = ":")
+    } else
+        "(missing #line/file info)"
+}
 
-RtangleRuncode <-  function(object, chunk, options)
+RtangleRuncode <- function(object, chunk, options)
 {
     if (!(options$engine %in% c("R", "S"))) return(object)
 
@@ -732,23 +739,18 @@ RtangleRuncode <-  function(object, chunk, options)
     } else
         chunkout <- object$output
 
-    if (object$annotate) {
-        lnos <- grep("^#line ", chunk, value = TRUE)
-        if(length(lnos)) {
-            srclines <- attr(chunk, "srclines")
- ##         srcfilenum <- attr(chunk, "srcFilenum")
-            ## this currently includes the chunk header
-            lno <- if (length(srclines)) paste(min(srclines), max(srclines), sep = "-") else srclines
-            fn <- sub('[^"]*"([^"]+).*', "\\1", lnos[1L])
-        }
-        cat("###################################################\n",
-            "### code chunk number ", options$chunknr,
-            ": ",
-            if(!is.null(options$label)) options$label
-            else paste(fn, lno, sep = ":"),
-            ifelse(options$eval, "", " (eval = FALSE)"), "\n",
-            "###################################################\n",
-            file = chunkout, sep = "")
+    showOut <- options$eval || !object$drop.evalFALSE
+    if(showOut) {
+        annotate <- object$annotate
+        if (is.logical(annotate) && annotate) {
+            cat("###################################################\n",
+                "### code chunk number ", options$chunknr, ": ",
+                if(!is.null(ol <- options$label)) ol else .RtangleCodeLabel(chunk),
+                if(!options$eval) " (eval = FALSE)", "\n",
+                "###################################################\n",
+                file = chunkout, sep = "")
+        } else if(is.function(annotate))
+            annotate(options, chunk = chunk, output = chunkout)
     }
 
     ## The next returns a character vector of the logical options
@@ -758,10 +760,12 @@ RtangleRuncode <-  function(object, chunk, options)
         cat("getOption(\"SweaveHooks\")[[\"", k, "\"]]()\n",
             file = chunkout, sep = "")
 
-    if (!options$show.line.nos)
-        chunk <- grep("^#line ", chunk, value = TRUE, invert = TRUE)
-    if (!options$eval) chunk <- paste("##", chunk)
-    cat(chunk, "\n", file = chunkout, sep = "\n")
+    if(showOut) {
+        if (!options$show.line.nos) # drop "#line ...." lines
+            chunk <- grep("^#line ", chunk, value = TRUE, invert = TRUE)
+        if (!options$eval) chunk <- paste("##", chunk)
+        cat(chunk, "\n", file = chunkout, sep = "\n")
+    }
     if (is.null(options$label) && options$split) close(chunkout)
     object
 }
